%%%----------------------------------------------------------------------
%%% File    : increment.erl
%%% Author  : Scott Lystig Fritchie <scott@sendmail.com>
%%% Purpose : A brute-force attempt to show how quickly Erlang's VM can
%%%           handle a couple of processes busily chatting with each other.
%%%----------------------------------------------------------------------

-module(increment).
-compile([verbose, report_errors, report_warnings, trace]).
-behaviour(gen_server).

-define(NAME, ?MODULE).
-define(Timeout, infinity).

%% External exports
-export([start_link/1]).
-export([get1/1, get_many/2, reset/2]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
       code_change/3]).
%% Internal usage only
-export([forever/0]).

%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
start_link(ArgList) ->
    io:format("XXX ~w:start_link ArgList = ~w\n", [?MODULE, ArgList]),
    case lists:keysearch(register_as, 1, ArgList) of
	{value, {register_as, Name}} ->
	    gen_server:start_link({local, Name}, ?MODULE, ArgList, []);
	_ ->
	    {error, "'register_as' tuple not found in ArgList"}
    end.

get1(Server) ->
    Timeout = ?Timeout,
    gen_server:call(Server, {get1}, Timeout).

get_many(Server, Total) ->
    get_many(Server, 0, Total, now()).
get_many(Server, Total, Total, {_, StartS, StartMs}) ->
    %% Ignoring Megaseconds not good but it's simple....
    {_, FinS, FinMs} = now(),
    Finish = FinS * 1000000 + FinMs,
    Start = StartS * 1000000 + StartMs,
    io:format("Start = ~w, Finish = ~w\n", [Start, Finish]),
    io:format("Looped ~w times in ~w secs, ~w usec/loop\n",
             [Total, (Finish - Start) / 1000000, (Finish - Start) / Total]),
    ok;
get_many(Server, N, Total, Start) ->
    get1(Server),
    get_many(Server, N + 1, Total, Start).

reset(Server, N) ->
    Timeout = ?Timeout,
    gen_server:call(Server, {reset, N}, Timeout).

%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init(ArgList) ->
    io:format("XXX ~w:init ArgList = ~w\n", [?MODULE, ArgList]),
    %% 
    %% We aren't doing anything with ArgList.  However, a real
    %% server probably would.
    %% 

    %% Spawn a few children, just to make appmon's display look pretty
    spawn_link(timer, sleep, [15000]),
    spawn_link(timer, sleep, [999000]),
    spawn_link(?MODULE, forever, []),
    
    {ok, 0}.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_call({get1}, From, State) ->
    {reply, State, State + 1};

handle_call({reset, N}, From, State) ->
    {reply, State, N}.

%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_cast(Msg, State) ->
    error_logger:info_msg("~w: ~s:handle_cast got ~w\n",
                         [self(), ?MODULE, Msg]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_info(Info, State) ->
    error_logger:info_msg("~w: ~s:handle_info got ~w\n",
                         [self(), ?MODULE, Info]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(Reason, State) ->
    ok.

%%----------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Preserve server state across code upgrades
%% Returns: {ok, NewState}
%%----------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    {ok, State}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

forever() ->
    receive
	i_will_never_get_this_silly_message ->
	    ok
    end.
